__all__ = ['parse_date', 'parse_datetime']

from typing import Any
import datetime


def parse_date(value: Any) -> datetime.date:
    if type(value) is datetime.date:
        return value
    elif type(value) is datetime.datetime:
        return value.date
    elif isinstance(value, str):
        return datetime.date.fromisoformat(value)
    return value
    
    
def parse_datetime(value: Any) -> datetime.datetime:
    if type(value) is datetime.date:
        return datetime.datetime(value.year, value.month, value.date)
    elif type(value) is datetime.datetime:
        return value
    elif isinstance(value, str):
        return datetime.datetime.fromisoformat(value)
    return value


__all__ = [
        'TypeHint',
        'parser',
        'Parser'
]


from .._imports import *
from .null import *
from dtfield.base_enum import *


JsonPrimitive = Union[str, float, int, bool, None]
DetaData = Union[dict, list, str, float, int, bool]
DetaKey = Union[str, None]
ExpireIn = Union[str, None]
ExpireAt = Union[datetime.datetime, int, float, None]
JsonSequence = list[JsonPrimitive]
JsonDict = dict[str, Union[JsonSequence, JsonPrimitive]]
Jsonable = Union[JsonDict, JsonSequence, JsonPrimitive]
DetaQuery = Union[dict[str, JsonPrimitive], list[dict[str, JsonPrimitive]]]
ExistParams = Union[list[str], str]
SearchParam = str
TypeArgs = tuple[type]
Types = Union[type, TypeArgs]
TupleSequenceTypes = (tuple, list, set, deque, UserList)
TupleMapTypes = (dict, UserDict, ChainMap)
ParserEntry = TypeVar('ParserEntry')
ParserResult = TypeVar('ParserResult')
ParserReturn = Union[ParserEntry, ParserResult]
GenericType = TypeVar('GenericType')
Number = TypeVar('Number', float, int, Decimal)


class ParserProtocol(Protocol):
    def __call__(self, value: ParserEntry) -> Union[ParserEntry, ParserResult]:
        ...


class ValidatorProtocol(Protocol):
    def __call__(self, value: GenericType, types: Types) -> GenericType:
        ...


class Parser(ABC):
    @classmethod
    def get(cls, value: ParserEntry, cast_type: type[ParserResult]):
        return cls.process(value, cast_type)

    @classmethod
    def process(cls, value: ParserEntry, cast_type: type[ParserResult]) -> ParserReturn:
        cast_analysis = TypeHint(cast_type)
        if cast_analysis.check_type(value):
            return value
        try:
            if cast_analysis.is_simple_type:
                if isinstance(value, cast_type):
                    return value
                elif is_null(value):
                    return ''
                return cls.simple_engine(cast_type)(value)
            elif cast_analysis.is_sequence:
                return cls.sequence_engine(value, cast_analysis.origin, cast_analysis.args)
            elif cast_analysis.is_mapping:
                return cls.mapping_engine(value, cast_analysis.origin, cast_analysis.args)
            else:
                return cls.generic_engine(value, cast_type)
        except ValueError as e:
            print(e)
            return value

    @classmethod
    def simple_engine(cls, cast_type: type[ParserResult]) -> ParserProtocol:
        if cast_type is str:
            return cls.str_engine
        elif cast_type is int:
            return cls.int_engine
        elif cast_type is float:
            return cls.float_engine
        elif cast_type is datetime.date:
            return cls.date_engine
        elif cast_type is datetime.datetime:
            return cls.datetime_engine
        elif cast_type is bytes:
            return cls.bytes_engine
        elif cast_type is Decimal:
            return cls.decimal_engine
        elif cast_type is list:
            return cls.list_engine
        elif cast_type is set:
            return cls.set_engine
        elif cast_type is tuple:
            return cls.tuple_engine
        elif hasattr(cast_type, 'CTXVAR'):
            return cls.str_engine
        elif issubclass(cast_type, (BaseEnum, Enum)):
            return lambda value: cls.enum_engine(value, cast_type)
        else:
            return cast_type

    @classmethod
    def sequence_engine(cls, value: ParserEntry, origin: type, args: tuple[type]) -> ParserReturn:
        if origin is list:
            if isinstance(value, (list, tuple, set)):
                return [cls.get(item, args[0]) for item in value]
            else:
                return [cls.get(value, args[0])]
        elif origin in [set, tuple]:
            if isinstance(value, (list, tuple, set)):
                return origin([cls.get(item, args[0]) for item in value])
            else:
                return origin([cls.get(value, args[0])])

    @classmethod
    def mapping_engine(cls, value: ParserEntry, origin: Union[dict, ChainMap], args: tuple[type]) -> ParserReturn:
        if origin is dict:
            if isinstance(value, (ChainMap, dict)):
                return {cls.get(k, args[0]): cls.get(v, args[-1]) for k, v in value.items()}
        elif origin is ChainMap:
            if isinstance(value, (ChainMap, dict)):
                return ChainMap({cls.get(k, args[0]): cls.get(v, args[-1]) for k, v in value.items()})

    @classmethod
    def generic_engine(cls, value: ParserEntry, generic: ParserResult) -> ParserReturn:
        if generic is Number:
            return cls.number_engine(value)

    @classmethod
    def list_engine(cls, value: ParserEntry) -> ParserReturn:
        if isinstance(value, (set, tuple)):
            return [*value]
        return value if isinstance(value, list) else [value]

    @classmethod
    def set_engine(cls, value: ParserEntry) -> ParserReturn:
        if isinstance(value, (list, tuple)):
            return {**value}
        return value if isinstance(value, set) else {value}

    @classmethod
    def tuple_engine(cls, value: ParserEntry) -> ParserReturn:
        if isinstance(value, (list, set)):
            return tuple([*value])
        return value if isinstance(value, tuple) else tuple([value])

    @classmethod
    def str_engine(cls, value: ParserEntry) -> ParserReturn:
        return str(value)

    @classmethod
    def int_engine(cls, value):
        if isinstance(value, str):
            return int(float(re.sub(r',', '.', value)).__round__(0))
        elif isinstance(value, Decimal):
            return int(value)
        elif isinstance(value, (datetime.date, datetime.datetime)):
            return value.toordinal()
        return value
    
    @classmethod
    def context_model(cls, value: dict, model_type: type['ContextBase']):
        return model_type.safe_create(**value)
        
    @classmethod
    def float_engine(cls, value):
        if isinstance(value, str):
            return float(re.sub(r',', '.', value))
        elif isinstance(value, int):
            return float(value)
        elif isinstance(value, Decimal):
            return float(value)
        elif isinstance(value, bytes):
            return float(re.sub(r',', '.', value.decode('utf-8')))
        return value

    @classmethod
    def date_engine(cls, value: ParserEntry) -> ParserReturn:
        if isinstance(value, str):
            return datetime.date.fromisoformat(value)
        elif isinstance(value, datetime.datetime):
            return value.date()
        elif isinstance(value, int):
            return datetime.date.fromordinal(value)

    @classmethod
    def enum_engine(cls, value: ParserEntry, enum_type: type[BaseEnum]) -> ParserReturn:
        try:
            return enum_type[value]
        except BaseException as e:
            print(e)
            return enum_type(value)

    @classmethod
    def datetime_engine(cls, value: ParserEntry) -> ParserReturn:
        if isinstance(value, str):
            return datetime.datetime.fromisoformat(value)
        elif isinstance(value, datetime.date):
            return datetime.datetime.fromordinal(value.toordinal())
        elif isinstance(value, int):
            return datetime.datetime.fromordinal(value)

    @classmethod
    def bytes_engine(cls, value: ParserEntry) -> ParserReturn:
        return str(value).encode('utf-8')

    @classmethod
    def decimal_engine(cls, value: ParserEntry) -> ParserReturn:
        if isinstance(value, str):
            return Decimal(re.sub(r'\.,', '.', value))
        elif isinstance(value, float):
            return Decimal(str(value))
        elif isinstance(value, int):
            return Decimal(str(float(value)))

    @classmethod
    def number_engine(cls, value: ParserEntry) -> ParserReturn:
        if isinstance(value, str):
            if re.match(r'\.,', value):
                return cls.get(value, float)
            else:
                return cls.get(value, int)
        elif isinstance(value, (datetime.date, datetime.datetime)):
            return value.toordinal()




class TypeHint:
    def __init__(self, type_hint):
        self.type_hint = type_hint
        
    def __str__(self):
        return f'{type(self).__name__}({self.type_hint})'
    
    @property
    def args(self) -> tuple[type]:
        return tuple([t for t in get_args(self.type_hint) if isinstance(t, type) if t is not type(None)])
    
    @property
    def is_args_generic(self):
        return all([not isinstance(item, type) for item in self.args])
    
    @property
    def origin(self):
        return get_origin(self.type_hint)
    
    @property
    def is_sequence(self):
        return self.origin in TupleSequenceTypes
    
    @property
    def is_mapping(self):
        return self.origin in TupleMapTypes
    
    @property
    def is_generic(self):
        return not isinstance(self.origin, type)
    
    @property
    def is_simple_type(self) -> bool:
        return not self.args and isinstance(self.type_hint, type)
    
    @property
    def is_optional(self):
        return type(None) in self.args
    
    @property
    def expected_type(self):
        if self.is_simple_type:
            return self.type_hint
        elif self.is_mapping or self.is_sequence:
            return self.origin
        else:
            return self.args
    
    @property
    def is_dict_model(self):
        return issubclass(self.type_hint, UserDict)
    
    @property
    def is_dataclass(self):
        return is_dataclass(self.type_hint)

    @property
    def is_string_list(self):
        return self.type_hint == list[str]
    
    def check_type(self, value: Any) -> bool:
        if self.is_simple_type:
            if self.is_dataclass:
                return isinstance(value, self.type_hint)
            elif self.is_dict_model:
                return isinstance(value, dict)
            elif self.is_string_list:
                return isinstance(value, list) and all([isinstance(item, str) for item in value])
            return isinstance(value, self.type_hint)
        elif self.is_generic:
            return isinstance(value, self.args)
        elif self.is_sequence:
            return isinstance(value, self.origin) and all([isinstance(item, self.args) for item in value])
        elif self.is_mapping:
            return isinstance(value, self.origin) and all(
                    [isinstance(item, tuple(self.args[1:])) for item in value.values()]) and all(
                    [isinstance(item, self.args[0]) for item in value.keys()])
        else:
            return False
        
    
    def parse(self, value: ParserEntry) -> ParserResult:
        if self.check_type(value):
            return value
        return Parser.get(value, self.type_hint)
            

def parser(value, type_hint):
    return Parser.get(value, type_hint)


__all__ =[
        'datetime',
        'create_task_group',
        'Enum',
        'Union',
        'Optional',
        'Any',
        'Callable',
        'TypeVar',
        'Iterable',
        'Type',
        'NamedTuple',
        'UserString',
        'UserDict',
        'UserList',
        'deque',
        'defaultdict',
        'ChainMap',
        'Context',
        'ContextVar',
        'copy_context',
        'wraps',
        'partialmethod',
        'cache',
        'partial',
        'Mapping',
        'get_origin',
        'get_type_hints',
        'get_args',
        'Self',
        'dataclass',
        'field',
        'fields',
        'MISSING',
        'Field',
        'is_dataclass',
        'Decimal',
        'ClassVar',
        'InitVar',
        'ABC',
        'abstractmethod',
        'groupby',
        'json',
        'asyncio',
        'uvicorn',
        'Response',
        'HTMLResponse',
        'JSONResponse',
        'RedirectResponse',
        'Request',
        're',
        'asdict',
        'astuple',
        'Protocol'
]

import datetime
import json
import asyncio
import re
import uvicorn
from enum import Enum
from abc import ABC, abstractmethod
from typing import Union, Optional, Any, Callable, TypeVar, Iterable, Type, NamedTuple, Mapping, get_type_hints, \
    get_args, get_origin, ClassVar, Protocol
from collections import UserString, UserDict, UserList, deque, defaultdict, ChainMap
from contextvars import Context, ContextVar, copy_context
from functools import wraps, partialmethod, cache, partial
from itertools import groupby
from dataclasses import fields, MISSING, field, Field, is_dataclass, dataclass, InitVar, asdict, astuple
from decimal import Decimal
from anyio import create_task_group
from typing_extensions import Self
from starlette.requests import Request
from starlette.responses import Response, HTMLResponse, JSONResponse, RedirectResponse




__all__ = ['parse_int', 'parse_float', 'parse_decimal']

import re
from typing import Any
from decimal import Decimal
import datetime


def parse_int(value: Any) -> int:
    if isinstance(value, str):
        return int(float(re.sub(r',', '.', value)).__round__(0))
    elif isinstance(value, float):
        return int(value)
    elif isinstance(value, (datetime.datetime, datetime.date)):
        return value.toordinal()
    elif isinstance(value, bool):
        return int(value)
    return value


def parse_float(value: Any) -> float:
    if isinstance(value, str):
        return float(re.sub(r',', '.', value)).__round__(2)
    elif isinstance(value, int):
        return float(value)
    elif isinstance(value, (datetime.datetime, datetime.date)):
        return float(value.toordinal())
    elif isinstance(value, bool):
        return float(value)
    return value

def parse_decimal(value: Any) -> Decimal:
    if isinstance(value, str):
        return Decimal(re.sub(r',', '.', value))
    elif isinstance(value, int):
        return Decimal(str(float(value)))
    elif isinstance(value, (datetime.datetime, datetime.date)):
        return Decimal(value.toordinal())
    elif isinstance(value, bool):
        return Decimal(value)
    return value


__all__ = ['BaseEnum']


from dtfield import Enum

class BaseEnum(Enum):
    """Enum base model class"""
    
    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.name == other.name
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, type(self)):
            return self.value < other.value
        return NotImplemented
    
    @classmethod
    def table(cls):
        return cls.__name__
    
    def json(self):
        return self.name
    
    @property
    def key(self):
        return self.name
    
    def __str__(self):
        return self.value
    
    @property
    def display(self):
        return self.value
    
    @classmethod
    def members(cls):
        return cls.__members__.values()
    
    @classmethod
    def option(cls, item: 'BaseEnum' = None, selected: bool = False):
        if not item:
            return '<option></option>'
        return f'<option id="{type(item).__name__}.{item.key}" value="{item.key}" ' \
               f'{"selected" if selected is True else ""}>{item.display}</option>'
    
    @classmethod
    def options(cls, default: str = None):
        if default:
            if isinstance(default, cls):
                default = default.name
        return ''.join([cls.option(), *[cls.option(member, member.key == default) for member in cls.members()]])




__all__ = [
        'json_dumps',
        'json_parse',
        'json_loads',
        'JsonEncoder'
]

# import json
# import datetime
# from enum import Enum
# from dataclasses import asdict, is_dataclass
# from collections import UserString, UserList, UserDict
# from typing import Any
from dtfield._imports import *
from dtbase import Jsonable


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()[:16]
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, Enum):
            return obj.name
        elif isinstance(obj, bytes):
            return obj.decode(encoding='utf-8')
        elif isinstance(obj, dict):
            return json_parse(obj)
        elif isinstance(obj, (list, tuple)):
            return [json_parse(item) for item in obj]
        elif isinstance(obj, UserString):
            return str(obj)
        elif isinstance(obj, (UserDict, UserList)):
            return json_parse(obj.data)
        elif hasattr(obj, 'asjson'):
            return obj.asjson()
        elif is_dataclass(obj) and not isinstance(obj, type):
            return json_parse(asdict(obj))
        return json.JSONEncoder.default(self, obj)


def json_loads(obj: str) -> Jsonable:
    return json.loads(obj)


def json_dumps(obj: Any) -> str:
    return json.dumps(obj, cls=JsonEncoder, indent=2, ensure_ascii=False)


def json_parse(obj: Any) -> Jsonable:
    return json_loads(json_dumps(obj))




__all__ = ['parse_str', 'parse_bytes']

from typing import Any
import datetime


def parse_str(value: Any) -> str:
    if isinstance(value, (datetime.datetime, datetime.date)):
        return value.isoformat()
    elif isinstance(value, bytes):
        return value.decode('utf-8')
    elif hasattr(value, 'key'):
        return value.key
    return str(value)


def parse_bytes(value: Any) -> bytes:
    if isinstance(value, (datetime.datetime, datetime.date)):
        return value.isoformat().encode('utf-8')
    elif hasattr(value, 'key'):
        return str(value.key).encode('utf-8')
    return str(value).encode('utf-8')


from __future__ import annotations

__all__ = [
        'ModelMap',
        'context_model',
        'BaseDescriptor',
        'BaseValidator',
        'ContextBase',
        'Descriptor',
        'Validator',
        'DescriptorModel',
        'compare_descriptors',
        'descriptors',
        'descriptors_items',
        'compare_dict',
        'compare_tuple',
        'hash_tuple',
        'hash_dict',
        'hash_descriptors',
        'check_hash',
        'check_compare',
        'instance_hash',
        'lesser_than',
        'orderby',
        'ordered',
        'reversed_dict'
]


from deta.base import FetchResponse
from dtfield._imports import *
from dtbase import *
from .parse import *
from .functions import *


context = copy_context()


def get_attr_or_item(obj: Any, name: str = 'key'):
    if isinstance(obj, Mapping):
        return obj.get(name)
    elif hasattr(obj, name):
        return getattr(obj, 'key')
    raise AttributeError(f'O objeto "{obj}" não possuir uma chave ou attributo "{name}".')


def make_dict(
        iterable: Iterable[T],
        key_func: Callable[[T], str] = get_attr_or_item,
        value_func: Callable[[T], Any] = lambda x: x,
        condition: Callable[[T], bool] = lambda x: True
    ) -> dict[str, T]:
    return {key_func(item): value_func(item) for item in iterable if condition(item) is True}


def filter_dict(data: Mapping[str, T], condition: Callable[[T], bool]) -> dict[str, T]:
    return make_dict(data.items(), key_func=lambda x: x[0], value_func=lambda x: x[1], condition=condition)

normalize_lower_if_string = partial(lambda x: x if not isinstance(x, str) else normalize_lower(x))
descriptors = partial(lambda cls: cls.descriptors())
descriptors_items = partial(lambda x: descriptors(x).items())
compare_descriptors = partial(lambda x: filter_dict(descriptors(x), condition=lambda x: x[1].compare is True))
compare_dict = partial(lambda x: {k: normalize_lower_if_string(v.__get__(x)) for k, v in compare_descriptors(x).items()})
compare_tuple = partial(lambda x: tuple(compare_dict(x).values()))
check_compare = partial(lambda self, other: compare_tuple(self) == compare_tuple(other))
hash_descriptors = partial(lambda x: filter_dict(descriptors(x), condition=lambda x: x[1].hash is True))
hash_dict = partial(lambda x: {k: v.__get__(x) for k, v in hash_descriptors(x).items()})
hash_tuple = partial(lambda x: tuple(hash_dict(x).values()))
check_hash = partial(lambda self, other: hash(hash_tuple(self)) == hash(hash_tuple(other)))
instance_hash = partial(lambda self: hash(hash_tuple(self)))
reversed_dict = partial(lambda x: {k: x[k] for k in reversed(x.keys())})
lesser_than = partial(lambda self, other: compare_tuple(self) < compare_tuple(other))
greater_than = partial(lambda self, other: compare_tuple(self) > compare_tuple(other))
orderby = partial(lambda iterable, public_name: sorted(iterable, key=lambda item: normalize_lower_if_string(getattr(item, public_name))))

def repr_func(self: ContextBase):
    items = (f'{k}={getattr(self, v.public_name)!r}' for k, v in self.descriptors().items() if v.public == True)
    return '{}({})'.format(self.clsname(), ', '.join(items))

def display_items(self: ContextBase) -> list[tuple[str, Any]]:
    if self.display_properties():
        return [(i.split('.')[-1], getter(self, i)) for i in self.display_properties()]
    return []

def display_repr(self: ContextBase):
    items = display_items(self)
    if items:
        return ', '.join([f'{i[0]}={i[1]}' for i in items ])
    return None

def eq_function(self: T, other: T) -> bool:
    return check_hash(self, other)


def ordered(iterable: list[DescriptorModel]):
    result = iterable
    if iterable:
        item = iterable[0]
        keys = compare_descriptors(item).keys()
        for key in reversed(keys):
            result = orderby(result, key)
    return result


@dataclass
class DescriptorModel(ABC):
    
    def __post_init__(self):
        for d in self.post_init_descriptors().values():
            if is_null(getattr(self, d.public_name)):
                setattr(self, d.public_name, d.post_init_factory())
    
    @classmethod
    def initfields(cls) -> tuple[Field, ...]:
        return tuple([f for f in fields(cls) if f.init == True])
    
    @classmethod
    @cache
    def initfields_keys(cls) -> list[str]:
        return [i.name for i in cls.initfields()]
    
    @classmethod
    def post_init_descriptors(cls):
        return filter_dict(cls.descriptors(), lambda x: x[1].post_init_factory is not MISSING)
    
    @classmethod
    def fields(cls) -> tuple[Field, ...]:
        return fields(cls)
    
    @classmethod
    def fieldsmap(cls) -> dict[str, Field]:
        return make_dict(cls.fields(),  lambda x: get_attr_or_item(x, 'public_name'))

    
    @classmethod
    def safe_create(cls, *args, **kwargs) -> Self:
        return cls(*args, **{k: v for k, v in kwargs.items() if k in cls.initfields_keys()})
    
    @classmethod
    def dataclass_bases(cls):
        return tuple([b for b in cls.mro() if is_dataclass(b)])
    
    @classmethod
    def clsname(cls):
        return cls.__name__
    
    @classmethod
    def fullvars(cls):
        return {**ChainMap(*[vars(b) for b in cls.mro()])}
    
    @classmethod
    def descriptors(cls):
        return filter_dict(cls.fullvars(), condition=lambda x: isinstance(x[1], BaseDescriptor))
    
    def asjson(self):
        return json_parse(asdict(self))
        
        
@dataclass
class ContextBase(DescriptorModel):
    CTXVAR: ClassVar[ContextVar] = None
    DETA_QUERY: ClassVar[Optional[DetaQuery]] = None
    TABLE: ClassVar[Optional[str]] = None
    ITEM_NAME: ClassVar[Optional[str]] = None
    DISPLAY_PROPERTIES: ClassVar[Optional[list[str]]] = None
    SINGULAR: ClassVar[Optional[str]] = None
    PLURAL: ClassVar[Optional[str]] = None

    @classmethod
    def singular(cls):
        return cls.SINGULAR or cls.clsname()
    
    @classmethod
    def plural(cls):
        return cls.PLURAL or f'{cls.singular()}s'
    
    @classmethod
    def display_properties(cls):
        return cls.DISPLAY_PROPERTIES or list()
    
    @classmethod
    def item_name(cls):
        return cls.ITEM_NAME or slug(cls.table())
    
    @classmethod
    def table(cls):
        return cls.TABLE or cls.clsname()
    
    @classmethod
    def dependant_descriptors(cls):
        return {d.public_name: d for d in cls.descriptors().values() if d.has_dependants is True}
    
    @classmethod
    def model_dependants(cls, collection: list = None):
        return model_dependants(cls, collection or list())

    @classmethod
    def from_context(cls, key: str) -> Self:
        data = cls.get_context()[key]
        if data:
            return cls.safe_create(**data)
        return None
    
    @classmethod
    async def set_context(cls):
        context.run(cls.CTXVAR.set, make_dict(await cls.fetch_all()))
        
    @classmethod
    def get_context(cls):
        return context.get(cls.CTXVAR)

    
    @classmethod
    async def fetch_all(cls, query: DetaQuery = None) -> list[dict[str, Jsonable]]:
        return await DetaBase(cls.table()).fetch_all(query or cls.DETA_QUERY)
    
    @classmethod
    async def fetch(cls, query: DetaQuery = None, last: Optional[str] = None, limit: int = 1000) -> FetchResponse:
        return await DetaBase(cls.table()).fetch(query or cls.DETA_QUERY, last, limit)
    
    @classmethod
    async def update_context(cls):
        async with create_task_group() as tks:
            for item in cls.model_dependants():
                tks.start_soon(item.set_context)
                
    @classmethod
    def instances_list(cls) -> list[Self]:
        return ordered([cls.safe_create(**i) for i in cls.get_context().values()])
    



def model_dependants(obj: type[ContextBase], collection: list = None):
    collection = collection or list()
    for item in obj.dependant_descriptors().values():
        models = item.dependants
        for model in models:
            if not model in collection:
                collection = model_dependants(model, collection)
    if not obj in collection:
        collection.append(obj)
    return collection



ModelMap: ChainMap[str, type[ContextBase]] = ChainMap()


def context_model(cls: type[ContextBase]):
    @wraps(cls)
    def wrapper():
        cls.CTXVAR = ContextVar(f'{cls.__name__}Var')
        ModelMap[cls.__name__] = cls
        cls.__repr__ = partialmethod(repr_func)
        cls.__eq__ = partialmethod(check_hash)
        cls.__lt__ = partialmethod(lesser_than)
        cls.__gt__ = partialmethod(greater_than)
        return cls
    
    return wrapper()


class BaseDescriptor:
    BASE_NUMBER_TYPES: ClassVar[TupleOfTypes] = (int, float, Decimal)
    BASE_DATETIME_TYPES: ClassVar[TupleOfTypes] = (datetime.datetime, datetime.date)
    STRING_TYPES: ClassVar[TupleOfTypes]
    BASE_TYPES: ClassVar[TupleOfTypes] = tuple([str, *BASE_NUMBER_TYPES, *BASE_DATETIME_TYPES])
    
    def __init__(
            self,
            default: Any = MISSING,
            # callabes
            default_factory: Optional[Callable[[], Any]] = MISSING,
            post_init_factory: Optional[Callable[[], Any]] = MISSING,
            auto: Optional[Callable[[DescriptorModel], Any]] = MISSING,
            update_auto: Optional[Callable[[DescriptorModel], Any]] = MISSING,
            cls_auto: Optional[Callable[[Type[DescriptorModel]], Any]] = MISSING,
            post_parse: Optional[Callable[[Any], Any]] = MISSING,
            pre_parse: Optional[Callable[[Any], Any]] = MISSING,
            # bools
            private: bool = False,
            repr: bool = True,
            compare: bool = True,
            hash: Optional[bool] = None,
            frozen: Optional[bool] = None,
            db: bool = True,
            tablekey: bool = False,
            multiple: bool = False,
            # maps
            metadata: Optional[Mapping] = None,
            # strings
            label: Optional[str] = None,
            table: Optional[str] = None,
            item_name: Optional[str] = None,
            # lists
            tables: Optional[list[str]] = None,
    ):
        self._default = default
        self.private = private
        self.auto = auto
        self.update_auto = update_auto
        self.default_factory = default_factory
        self.post_init_factory = post_init_factory
        self.cls_auto = cls_auto
        self._label = label
        self._repr = repr
        self.compare = compare
        self.hash = hash
        self.metadata = metadata
        self.table = table
        self.tables = tables
        self.multiple = multiple
        self._item_name = item_name
        self.db = db
        self._tablekey = tablekey
        self.frozen = frozen
        self.post_parse = post_parse
        self.pre_parse = pre_parse
        
    def __repr__(self):
        return '{}({})'.format(
                type(self).__name__,
                ', '.join([f'{k}={getattr(self, k)!r}' for k in [
                        'field_type', 'public_name', 'required', 'owner', 'dependants', 'db'
                ]])
        )
    
    @property
    def public(self):
        return True if all([self.private == False, self.repr == True]) else False
    
    @property
    def has_dependants(self):
        return True if any([self.tables is not None, self.table is not None]) else False
    
    @property
    def tablekey(self):
        return all([self.has_dependants is True, any([self._tablekey is True, self.tables is not None])])
    
    @property
    def item_name(self):
        if self._item_name:
            return self._item_name
        elif self.table:
            return ModelMap.get(self.table).item_name()
        return re.sub(r'_key', '', self.public_name)
            
    @property
    def repr(self):
        if self.private:
            return False
        return self._repr
        
    @property
    def dependants(self) -> tuple[Optional[Type[ContextBase]]]:
        if self.tables:
            return tuple([ModelMap.get(item) for item in self.tables])
        elif self.table:
            return tuple([ModelMap.get(self.table)])
        return tuple()
    
    @property
    def required(self):
        return True if all([
                self._default is MISSING,
                self.default_factory is MISSING,
                self.auto is MISSING,
                self.cls_auto is MISSING,
                self.update_auto is MISSING,
                self.post_init_factory is MISSING,
        ]) else False
    
    @property
    def default(self):
        if self._default is MISSING:
            if self.required is False:
                return None
        return self._default
    
    @property
    def label(self):
        return self._label or self.public_name
    
    @property
    def field_type(self):
        return get_type_hints(self.owner)[self.public_name]
    
    def __set_name__(self, owner, name):
        self.private_name = f'_{name}'
        self.public_name = name
        self.owner: Type[DescriptorModel] = owner
    
    def __get__(self, instance, owner=None):
        if instance is None:
            return self.default
        return getattr(instance, self.private_name)
    
    def __set__(self, instance, value):
        if self.frozen:
            if not is_null(value):
                value = getattr(instance, self.private_name, value)
        else:
            value = self.parse(instance, self.set_lookup(instance, value))
        self.validate(instance, value)
        setattr(instance, self.private_name, value)
        if not is_null(value):
            if self.has_dependants:
                if self.tablekey:
                    model = ModelMap.get(value.split('.')[0])
                    setattr(instance, self.item_name, model.from_context(value.split('.')[-1]))
                else:
                    model = ModelMap.get(self.table)
                    setattr(instance, self.item_name, model.from_context(value))

    
    def set_lookup(self, instance, value):
        if self.update_auto is not MISSING:
            value = self.update_auto(instance)
        else:
            if value is None:
                if self.auto is not MISSING:
                    value = self.auto(instance)
                elif self.cls_auto is not MISSING:
                    value = self.cls_auto(self.owner)
                if self.default_factory is not MISSING:
                    value = self.default_factory()
        return value
    
    def parse(self, instance, value):
        if all([is_null(value) is False, is_null(self.pre_parse) is False]):
            value = self.pre_parse(value)
        value = Parser.get(value, self.field_type)
        if all([is_null(value) is False, is_null(self.post_parse) is False]):
            value = self.post_parse(value)
        return value
        # print(self.expected_type)
        # if is_null(value):
        #     return ''
        # elif isinstance(value, self.expected_type):
        #     return value
        # elif self.type in self.STRING_TYPES:
        #     return parse_str(value) if self.type is str else parse_bytes(value)
        # elif self.type in self.BASE_DATETIME_TYPES:
        #     return parse_datetime(value) if self.type is datetime.datetime else parse_date(value)
        # elif self.type in self.BASE_NUMBER_TYPES:
        #     return parse_int(value) if self.type is int else parse_float(
        #         value) if self.type is float else parse_decimal(value)
        # return value
    
    def validate(self, instance, value):
        pass
    
    @property
    def name(self):
        return '{}.{}'.format(self.owner.__name__, self.public_name)
    
    @property
    def type_hint(self):
        return TypeHint(self.field_type)
    
    @property
    def expected_type(self):
        return self.type_hint.expected_type
        # if self.type in self.BASE_TYPES:
        #     return self.type
        # elif is_dataclass(self.type):
        #     return str
        # elif self.tables:
        #     return (i for i in [list, str])
        # elif isinstance(self.type, type):
        #     return self.type
        # elif get_args(self.type):
        #     if get_origin(self.type) in [Union]:
        #         return tuple([*get_args(self.type)])
        #     else:
        #         return get_origin(self.type)
        # return self.type


class BaseValidator(BaseDescriptor):
    def __init__(
            self,
            step: Optional[int, float] = None,
            min: Optional[int, float] = None,
            max: Optional[int, float] = None,
            min_lenght: Optional[int] = None,
            max_lenght: Optional[int] = None,
            predicate: Optional[Callable[[DescriptorModel], bool]] = None,
            **kwargs):
        self.min = min
        self.max = max
        self.step = step
        self.min_lenght = min_lenght
        self.max_lenght = max_lenght
        self.predicate = predicate
        super().__init__(**kwargs)
    
    def validate_null(self, instance, value):
        if is_null(value):
            if self.required:
                raise ValueError(f'{self.name} não pode ser nulo.')
    
    def validate_type(self, instance, value):
        if not self.type_hint.check_type(value):
            raise ValueError(
                f'{self.name} exige como tipo(s) "{self.expected_type}". O encontrado para "{value}" foi "{type(value)}".')
    
    def validate_other(self, instance, value):
        if self.field_type in self.BASE_NUMBER_TYPES:
            self.validate_min(instance, value)
            self.validate_max(instance, value)
        elif hasattr(self.field_type, '__len__'):
            self.validate_min_length(instance, value)
            self.validate_max_length(instance, value)
        self.validate_predicate(instance, value)
    
    def validate_min_length(self, instance, value):
        if self.min_lenght:
            if len(value) < self.min_lenght:
                raise ValueError(
                    f'{self.name} não pode ter tamanho menor que "{self.min_lenght}". O encontrado para "{value}" é "{len(value)}".')
    
    def validate_max_length(self, instance, value):
        if self.max_lenght:
            if len(value) > self.max_lenght:
                raise ValueError(
                    f'{self.name} não pode ter tamanho maior que "{self.max_lenght}". O encontrado para "{value}" é "{len(value)}".')
    
    def validate_min(self, instance, value):
        if self.min:
            if self.min > value:
                raise ValueError(f'{self.name} não pode ser menor que "{self.min}". O encontrado é "{value}".')
    
    def validate_max(self, instance, value):
        if self.max:
            if self.max < value:
                raise ValueError(f'{self.name} não pode ser maior que "{self.max}". O encontrado é "{value}".')
    
    def validate_predicate(self, instance, value):
        if self.predicate:
            if self.predicate(value) is False:
                raise ValueError(f'{self.name} não passou no teste de predicativo com o valor "{value}".')
    
    def validate(self, instance, value):
        self.validate_null(instance, value)
        if not is_null(value):
            self.validate_type(instance, value)
            self.validate_other(instance, value)


class Descriptor(BaseDescriptor):
    pass


class Validator(BaseValidator):
    pass





from __future__ import annotations

__all__ = [
        'slug',
        'normalize',
        'normalize_lower',
        'hyphen_to_underscore',
        'underscore_to_hyphen',
        'remove_extra_whitespaces',
        'underscore_text',
        'titlecase',
        'getter',
        'map_setter',
        'last',
        'first',
        'keyattr',
        'keyitem',
        'key'
    ]

import re
from typing import Any
from operator import getitem, attrgetter, itemgetter
from unidecode import unidecode


first = lambda x: getitem(x, 0)
last = lambda x: getitem(x, -1)
keyattr = attrgetter('key')
keyitem = itemgetter('key')
key = lambda x: keyitem(x) if isinstance(x, dict) else keyattr(x) if hasattr(x, 'key') else None


def map_setter(obj: dict) -> dict:
    for k, v in obj.items():
        setattr(obj, k, property(lambda obj: obj[k]))
    return obj
    

def getter(obj: Any, prop: str, default: Any = None) -> Any:
    k = (p for p in prop.strip().split('.') if p)
    value = getattr(obj, next(k), None)
    try:
        while True:
            value = getattr(value, next(k), None)
    finally:
        return value or default
    

def remove_extra_whitespaces(string: str) -> str:
    return re.sub(r'\s+', ' ', string.strip())

def normalize(string: str) -> str:
    return unidecode(string.strip())

def normalize_lower(string: str) -> str:
    return normalize(string).lower()

def separate_words(string: str) -> str:
    def analyse_low_up(value: str):
        result = re.search(r'([a-z][A-Z])', value)
        if result:
            mid = result.span()[0] + 1
            return analyse_low_up(value[:mid] + ' ' + value[mid:])
        return value
    
    def analyse_up_up_low(value: str):
        result = re.search(r'([A-Z][A-Z][a-z])', value)
        if result:
            mid = result.span()[-1] - 2
            return analyse_up_up_low(value[:mid] + ' ' + value[mid:])
        return value
    
    return remove_extra_whitespaces(analyse_up_up_low(analyse_low_up(string)))

def underscore_text(string: str):
    return re.sub(r'\s+|-+', '_', separate_words(string))


def slug(string: str) -> str:
    return normalize_lower(underscore_text(string))


def underscore_to_hyphen(string: str) -> str:
    return re.sub(r'_+', '-', string.strip())

def hyphen_to_underscore(string: str) -> str:
    return re.sub(r'-+', '_', string.strip())


def titlecase(string: str) -> str:
    return ' '.join([w.title() if not w.isupper() and not len(w) <=2 else w for w in separate_words(string).split()])


__all__ = ['is_null']

from typing import Any
from dataclasses import MISSING


def is_null(value: Any) -> bool:
    return True if any([
            value is None,
            value == '',
            value == b'',
            value == [],
            value == {},
            value == dict(),
            value == MISSING
    ]) else False
    


__all__ = [
        'T',
        'TupleOfTypes',
        'Key',
        'TableKey',
        'KeyList',
        'TableKeyList',
]

from typing import TypeVar

T = TypeVar('T')
TupleOfTypes = tuple[type]
Key = TypeVar('Key')
TableKey = TypeVar('TableKey')
KeyList = TypeVar('KeyList')
TableKeyList = TypeVar('TableKeyList')


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dtfield', 'dtfield.parse']

package_data = \
{'': ['*']}

install_requires = \
['Unidecode>=1.3.6,<2.0.0',
 'anyio>=3.7.0,<4.0.0',
 'dtbase>=0.0.4,<0.0.5',
 'typing-extensions>=4.6.3,<5.0.0',
 'uvicorn>=0.22.0,<0.23.0']

setup_kwargs = {
    'name': 'dtfield',
    'version': '0.1.6',
    'description': '',
    'long_description': None,
    'author': 'Daniel Arantes',
    'author_email': 'arantesdv@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
